// Synthesizes to 1006 4-LUTs and 2 MULT18X18s at 150 MHz in XC3Sxxx-4 using XST-10.1i

// One channel of cavity input processing for a vector-sum context
// "clk", "phase" and "mod2" inputs are common to all cavities
// "phase_shift" and "gain" inputs are used to adjust for RF input channel
//   (including cable) differences
// "iq_out" output is ready for summation across instantiations

// Resource usage compatible with fitting 8 channels on a US$40 XC3S1000,
// or 32 channels on a US$250 EP2C50 (or, of course, a US$350 EP2C70).

`timescale 1ns / 1ns

`ifdef FAKE_CORDIC
`include "constants.vams"
`endif

module cav_input(
	input clk,  // timespec 6.66 ns
	input  signed [13:0] cavity_in,
	input                mod2,    // Phase master for I/Q mux of output
	input         [16:0] phase,   // Digital "local oscillator"
	input         [16:0] phase_shift,
	input  signed [15:0] gain,
	output signed [15:0] iq_out
);

// phase for demodulator
// phase_shift overflow just wraps by 2*pi == 2^17
reg [16:0] phase0=0, phase1=0, phase2=0, phased=0;
always @(posedge clk) begin
	phase0 <= phase + phase_shift;
	phase1 <= phase0;
	phase2 <= phase1;
	phased <= mod2 ? phase2 : phase0;  // pipeline correct here?
end

// CORDIC to generate demodulator drive
wire signed [17:0] cosd, sind;
wire [16:0] dummyd;
`ifdef FAKE_CORDIC
// used in simulation only to get perfect results
assign cosd = $floor($cos(phased*`M_TWO_PI/131072.0)*gain*4*1.64676+0.5);
assign sind = $floor($sin(phased*`M_TWO_PI/131072.0)*gain*4*1.64676+0.5);
`else
cordicg c2(clk, 1'b0, gain, 16'b0, phased, cosd[17:2], sind[17:2], dummyd);
assign cosd[1:0]=0;
assign sind[1:0]=0;
`endif

// 1-z^{-2} filter
reg signed [13:0] cavity0=0;
reg signed [13:0] cavity1=0;
reg signed [13:0] cavity2=0;
reg signed [14:0] cav_nodc;
always @(posedge clk) begin
	cavity0 <= cavity_in;  // waste one cycle getting data into fabric
	cavity1 <= cavity0;
	cavity2 <= cavity1;
	cav_nodc <= cavity0 - cavity2;
end

// downconvert input signal to I and Q
downconvert d1(clk, cav_nodc, iq_out, cosd, sind, ~mod2);

endmodule
