// cordicg.v
// CORDIC processor
// Larry Doolittle, LBNL
// Synthesizes to 912 4-LUTs at 115 MHz in XC3Sxxx-4 using XST-10.1i

// 16 bit input for x, y
// 17 bit input for phase, including highest order bit (pi)
//  (but because of rounding errors, the output is really only accurate
//   to about 14 bits)
// still could use some language cleanup
// reference:
//   http://www.fpga-guru.com/cordic.htm

`timescale 1ns / 1ns

module addsubg(a,b,sum,control);
	parameter size=16;
	input [size-1:0] a, b;
	input control;
	output [size-1:0] sum;
	assign sum = control ? (a + b) : (a - b);
endmodule

module cstageg(clk, opin, xin, yin, zin, ain, opout, xout, yout, zout);
	parameter shift=1;
	parameter zwidth=16;
	parameter width=16;
	input clk;
	input opin;
	input [width-1:0] xin, yin;
	input [zwidth-1:0] zin, ain;
	output reg opout;
	output reg [width-1:0] xout, yout;
	output reg [zwidth-1:0] zout;

	wire control = opin ? yin[width-1] : zin[zwidth-1];
	wire [width-1:0] xv, yv;
	wire [zwidth-1:0] zv;
	addsubg #( width) ax(xin, {{(shift){yin[width-1]}},yin[width-1:shift]}, xv,  control);
	addsubg #( width) ay(yin, {{(shift){xin[width-1]}},xin[width-1:shift]}, yv, ~control);
	addsubg #(zwidth) az(zin,                                         ain , zv,  control);
	always @(posedge clk) begin
		opout <= opin;
		xout  <= xv;
		yout  <= yv;
		zout  <= zv;
	end
endmodule

module cordicg(
	input clk,   // timespec 8.69 ns
	input opin,  //  1 forces y to zero (rect to polar), 0 forces theta to zero (polar to rect).
	input [15:0] xin,
	input [15:0] yin,
	input [16:0] phasein,
	output [15:0] xout,
	output [15:0] yout,
	output [16:0] phaseout);

// input buffer stage (routing)
reg        opin0;
reg [15:0] xin0, yin0;
reg [16:0] phasein0;
always @(posedge clk) begin
	opin0    <= opin;
	xin0     <= xin;
	yin0     <= yin;
	phasein0 <= phasein;
end

// zero stage: doesn't quite fit the pattern
           wire [16:0] xw0, yw0 ; wire [17:0] zw0;
reg  op0 ; reg  [16:0] x0,  y0  ; reg  [17:0] z0;
wire control0 = opin0 ? ~xin0[15] : phasein0[16]^phasein0[15];
addsubg #(17) ax0 (17'd0, {xin0,1'b0}, xw0, ~control0);
addsubg #(17) ay0 (17'd0, {yin0,1'b0}, yw0, ~control0);
assign zw0 = {phasein0[16]^control0,phasein0[15:0],1'b0};
always @(posedge clk) begin op0 <= opin0; x0 <= xw0; y0 <= yw0; z0 <= zw0; end

// first stage: can't use cstage because repeat operator of zero is illegal
           wire [16:0] xw1, yw1 ; wire [17:0] zw1;
reg  op1 ; reg  [16:0] x1,  y1  ; reg  [17:0] z1;
wire control1 = op0 ? y0[16] : z0[16];
addsubg #(17) ax1 (x0, y0, xw1,  control1);
addsubg #(17) ay1 (y0, x0, yw1, ~control1);
addsubg #(18) az1 (z0, 18'd32768, zw1,  control1);
always @(posedge clk) begin op1 <= op0; x1 <= xw1; y1 <= yw1; z1 <= zw1; end

// cut and pasted output from cordicgx.m
wire op2 ; wire [16:0] x2 , y2 ; wire [17:0] z2 ;  cstageg #(  1, 18, 17) cs1  (clk, op1 , x1 ,  y1 , z1 , 18'd19344, op2 , x2 ,  y2 ,  z2 );
wire op3 ; wire [16:0] x3 , y3 ; wire [17:0] z3 ;  cstageg #(  2, 18, 17) cs2  (clk, op2 , x2 ,  y2 , z2 , 18'd10221, op3 , x3 ,  y3 ,  z3 );
wire op4 ; wire [16:0] x4 , y4 ; wire [17:0] z4 ;  cstageg #(  3, 18, 17) cs3  (clk, op3 , x3 ,  y3 , z3 , 18'd5188 , op4 , x4 ,  y4 ,  z4 );
wire op5 ; wire [16:0] x5 , y5 ; wire [17:0] z5 ;  cstageg #(  4, 18, 17) cs4  (clk, op4 , x4 ,  y4 , z4 , 18'd2604 , op5 , x5 ,  y5 ,  z5 );
wire op6 ; wire [16:0] x6 , y6 ; wire [17:0] z6 ;  cstageg #(  5, 18, 17) cs5  (clk, op5 , x5 ,  y5 , z5 , 18'd1303 , op6 , x6 ,  y6 ,  z6 );
wire op7 ; wire [16:0] x7 , y7 ; wire [17:0] z7 ;  cstageg #(  6, 18, 17) cs6  (clk, op6 , x6 ,  y6 , z6 , 18'd652  , op7 , x7 ,  y7 ,  z7 );
wire op8 ; wire [16:0] x8 , y8 ; wire [17:0] z8 ;  cstageg #(  7, 18, 17) cs7  (clk, op7 , x7 ,  y7 , z7 , 18'd326  , op8 , x8 ,  y8 ,  z8 );
wire op9 ; wire [16:0] x9 , y9 ; wire [17:0] z9 ;  cstageg #(  8, 18, 17) cs8  (clk, op8 , x8 ,  y8 , z8 , 18'd163  , op9 , x9 ,  y9 ,  z9 );
wire op10; wire [16:0] x10, y10; wire [17:0] z10;  cstageg #(  9, 18, 17) cs9  (clk, op9 , x9 ,  y9 , z9 , 18'd81   , op10, x10,  y10,  z10);
wire op11; wire [16:0] x11, y11; wire [17:0] z11;  cstageg #( 10, 18, 17) cs10 (clk, op10, x10,  y10, z10, 18'd41   , op11, x11,  y11,  z11);
wire op12; wire [16:0] x12, y12; wire [17:0] z12;  cstageg #( 11, 18, 17) cs11 (clk, op11, x11,  y11, z11, 18'd20   , op12, x12,  y12,  z12);
wire op13; wire [16:0] x13, y13; wire [17:0] z13;  cstageg #( 12, 18, 17) cs12 (clk, op12, x12,  y12, z12, 18'd10   , op13, x13,  y13,  z13);
wire op14; wire [16:0] x14, y14; wire [17:0] z14;  cstageg #( 13, 18, 17) cs13 (clk, op13, x13,  y13, z13, 18'd5    , op14, x14,  y14,  z14);
wire op15; wire [16:0] x15, y15; wire [17:0] z15;  cstageg #( 14, 18, 17) cs14 (clk, op14, x14,  y14, z14, 18'd3    , op15, x15,  y15,  z15);
wire op16; wire [16:0] x16, y16; wire [17:0] z16;  cstageg #( 15, 18, 17) cs15 (clk, op15, x15,  y15, z15, 18'd1    , op16, x16,  y16,  z16);

assign xout = x16[16:1];
assign yout = y16[16:1];
assign phaseout = z16[17:1];

endmodule
