// Synthesizes to 65 4-LUTs at 120 MHz in XC3Sxxx-4 using XST-10.1i

`timescale 1ns / 1ns

module rat_dds(
	input clk,   // timespec 8.33 ns
	output reg [16:0] ang,
	input [19:0] phase_step_h,
	input [11:0] phase_step_l,
	input [11:0] modulo
);

// 12-bit modulo supports largest known periodicity in a
// suggested LLRF system, 1427 for JLab.  For more normal
// periodicities, use a multiple to get finer granularity.
// Note that the downloaded modulo control is the 2's complement
// of the mathematical modulus.  This is on purpose, so a
// power-on default of 0 corresponds to modulus 4096, i.e.,
// it is a traditional 32-bit binary DDS.
//
// No reset, sorry.  It would be easy to add if you want it.
//
// Example setups:
//
// SSRF IF/F_s ratio 8/11, use
//     modulo = 4096 - 372*11 = 4
//     phase_step_h = 2^20*8/11 = 762600
//     phase_step_l = (2^20*8%11)*372 = 2976
//
// Argonne RIA test IF/F_s ratio 9/13, use
//     modulo = 4096 - 315*13 = 1
//     phase_step_h = 2^20*9/13 = 725937
//     phase_step_l = (2^20*9%13)*315 = 945
//
// Fermilab ILC IF/F_s ratio 21/101, use
//     modulo = 4096 - 40*101 = 56
//     phase_step_h = 2^20*21/101 = 218020
//     phase_step_l = (2^20*21%101)*40 = 3040
//
// See rot_dds_config

reg carry=0;
reg [19:0] phase_h=0, phase_step_hp=0;
reg [11:0] phase_l=0;
always @(posedge clk) begin
	{carry, phase_l} <= (carry ? modulo : 0) + phase_l + phase_step_l;
	phase_step_hp <= phase_step_h;
	phase_h <= phase_h + phase_step_hp + carry;
	ang <= phase_h[19:3];
end

endmodule
